@@ -32,12 +32,6 @@
  * %End-Header%
  */
 
-/*
- * Force inclusion of SVID stuff since we need it if we're compiling in
- * gcc-wall wall mode
- */
-#define _SVID_SOURCE
-
 #ifdef _WIN32
 #define _WIN32_WINNT 0x0500
 #include <windows.h>
