@@ -1,7 +1,7 @@
 /*
  * findmnt(8)
  *
- * Copyright (C) 2010,2011 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2010-2014 Red Hat, Inc. All rights reserved.
  * Written by Karel Zak <kzak@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -36,12 +36,12 @@
 # include <libudev.h>
 #endif
 #include <libmount.h>
+#include <libsmartcols.h>
 
 #include "pathnames.h"
 #include "nls.h"
 #include "closestream.h"
 #include "c.h"
-#include "tt.h"
 #include "strutils.h"
 #include "xalloc.h"
 #include "optutils.h"
@@ -58,7 +58,17 @@ enum {
 	FL_SUBMOUNTS	= (1 << 8),
 	FL_POLL		= (1 << 9),
 	FL_DF		= (1 << 10),
-	FL_ALL		= (1 << 11)
+	FL_ALL		= (1 << 11),
+	FL_UNIQ		= (1 << 12),
+	FL_BYTES	= (1 << 13),
+	FL_NOCACHE	= (1 << 14),
+
+	/* basic table settings */
+	FL_ASCII	= (1 << 20),
+	FL_RAW		= (1 << 21),
+	FL_NOHEADINGS	= (1 << 22),
+	FL_EXPORT	= (1 << 23),
+	FL_TREE		= (1 << 24)
 };
 
 /* column IDs */
@@ -87,9 +97,7 @@ enum {
 	COL_OPT_FIELDS,
 	COL_PROPAGATION,
 	COL_FREQ,
-	COL_PASSNO,
-
-	FINDMNT_NCOLUMNS
+	COL_PASSNO
 };
 
 enum {
@@ -102,48 +110,63 @@ enum {
 struct colinfo {
 	const char	*name;		/* header */
 	double		whint;		/* width hint (N < 1 is in percent of termwidth) */
-	int		flags;		/* tt flags */
+	int		flags;		/* libsmartcols flags */
 	const char      *help;		/* column description */
 	const char	*match;		/* pattern for match_func() */
 	void		*match_data;	/* match specific data */
 };
 
 /* columns descriptions (don't use const, this is writable) */
-static struct colinfo infos[FINDMNT_NCOLUMNS] = {
-	[COL_SOURCE]       = { "SOURCE",       0.25, TT_FL_NOEXTREMES, N_("source device") },
-	[COL_TARGET]       = { "TARGET",       0.30, TT_FL_TREE | TT_FL_NOEXTREMES, N_("mountpoint") },
-	[COL_FSTYPE]       = { "FSTYPE",       0.10, TT_FL_TRUNC, N_("filesystem type") },
-	[COL_OPTIONS]      = { "OPTIONS",      0.10, TT_FL_TRUNC, N_("all mount options") },
-	[COL_VFS_OPTIONS]  = { "VFS-OPTIONS",  0.20, TT_FL_TRUNC, N_("VFS specific mount options") },
-	[COL_FS_OPTIONS]   = { "FS-OPTIONS",   0.10, TT_FL_TRUNC, N_("FS specific mount options") },
+static struct colinfo infos[] = {
+	[COL_SOURCE]       = { "SOURCE",       0.25, SCOLS_FL_NOEXTREMES, N_("source device") },
+	[COL_TARGET]       = { "TARGET",       0.30, SCOLS_FL_TREE| SCOLS_FL_NOEXTREMES, N_("mountpoint") },
+	[COL_FSTYPE]       = { "FSTYPE",       0.10, SCOLS_FL_TRUNC, N_("filesystem type") },
+	[COL_OPTIONS]      = { "OPTIONS",      0.10, SCOLS_FL_TRUNC, N_("all mount options") },
+	[COL_VFS_OPTIONS]  = { "VFS-OPTIONS",  0.20, SCOLS_FL_TRUNC, N_("VFS specific mount options") },
+	[COL_FS_OPTIONS]   = { "FS-OPTIONS",   0.10, SCOLS_FL_TRUNC, N_("FS specific mount options") },
 	[COL_LABEL]        = { "LABEL",        0.10, 0, N_("filesystem label") },
 	[COL_UUID]         = { "UUID",           36, 0, N_("filesystem UUID") },
 	[COL_PARTLABEL]    = { "PARTLABEL",    0.10, 0, N_("partition label") },
 	[COL_PARTUUID]     = { "PARTUUID",       36, 0, N_("partition UUID") },
 	[COL_MAJMIN]       = { "MAJ:MIN",         6, 0, N_("major:minor device number") },
-	[COL_ACTION]       = { "ACTION",         10, TT_FL_STRICTWIDTH, N_("action detected by --poll") },
-	[COL_OLD_OPTIONS]  = { "OLD-OPTIONS",  0.10, TT_FL_TRUNC, N_("old mount options saved by --poll") },
+	[COL_ACTION]       = { "ACTION",         10, SCOLS_FL_STRICTWIDTH, N_("action detected by --poll") },
+	[COL_OLD_OPTIONS]  = { "OLD-OPTIONS",  0.10, SCOLS_FL_TRUNC, N_("old mount options saved by --poll") },
 	[COL_OLD_TARGET]   = { "OLD-TARGET",   0.30, 0, N_("old mountpoint saved by --poll") },
-	[COL_SIZE]         = { "SIZE",            5, TT_FL_RIGHT, N_("filesystem size") },
-	[COL_AVAIL]        = { "AVAIL",           5, TT_FL_RIGHT, N_("filesystem size available") },
-	[COL_USED]         = { "USED",            5, TT_FL_RIGHT, N_("filesystem size used") },
-	[COL_USEPERC]      = { "USE%",            3, TT_FL_RIGHT, N_("filesystem use percentage") },
-	[COL_FSROOT]       = { "FSROOT",       0.25, TT_FL_NOEXTREMES, N_("filesystem root") },
-	[COL_TID]          = { "TID",             4, TT_FL_RIGHT, N_("task ID") },
-	[COL_ID]           = { "ID",              2, TT_FL_RIGHT, N_("mount ID") },
-	[COL_OPT_FIELDS]   = { "OPT-FIELDS",   0.10, TT_FL_TRUNC, N_("optional mount fields") },
+	[COL_SIZE]         = { "SIZE",            5, SCOLS_FL_RIGHT, N_("filesystem size") },
+	[COL_AVAIL]        = { "AVAIL",           5, SCOLS_FL_RIGHT, N_("filesystem size available") },
+	[COL_USED]         = { "USED",            5, SCOLS_FL_RIGHT, N_("filesystem size used") },
+	[COL_USEPERC]      = { "USE%",            3, SCOLS_FL_RIGHT, N_("filesystem use percentage") },
+	[COL_FSROOT]       = { "FSROOT",       0.25, SCOLS_FL_NOEXTREMES, N_("filesystem root") },
+	[COL_TID]          = { "TID",             4, SCOLS_FL_RIGHT, N_("task ID") },
+	[COL_ID]           = { "ID",              2, SCOLS_FL_RIGHT, N_("mount ID") },
+	[COL_OPT_FIELDS]   = { "OPT-FIELDS",   0.10, SCOLS_FL_TRUNC, N_("optional mount fields") },
 	[COL_PROPAGATION]  = { "PROPAGATION",  0.10, 0, N_("VFS propagation flags") },
-	[COL_FREQ]         = { "FREQ",            1, TT_FL_RIGHT, N_("dump(8) period in days [fstab only]") },
-	[COL_PASSNO]       = { "PASSNO",          1, TT_FL_RIGHT, N_("pass number on parallel fsck(8) [fstab only]") }
+	[COL_FREQ]         = { "FREQ",            1, SCOLS_FL_RIGHT, N_("dump(8) period in days [fstab only]") },
+	[COL_PASSNO]       = { "PASSNO",          1, SCOLS_FL_RIGHT, N_("pass number on parallel fsck(8) [fstab only]") }
 };
 
+/* columns[] array specifies all currently wanted output column. The columns
+ * are defined by infos[] array and you can specify (on command line) each
+ * column twice. That's enough, dynamically allocated array of the columns is
+ * unnecessary overkill and over-engineering in this case */
+static int columns[ARRAY_SIZE(infos) * 2];
+static int ncolumns;
+
+static inline size_t err_columns_index(size_t arysz, size_t idx)
+{
+	if (idx >= arysz)
+		errx(EXIT_FAILURE, _("too many columns specified, "
+				     "the limit is %zu columns"),
+				arysz - 1);
+	return idx;
+}
+
+#define add_column(ary, n, id)	\
+		((ary)[ err_columns_index(ARRAY_SIZE(ary), (n)) ] = (id))
+
 /* global flags */
 static int flags;
-static int tt_flags;
 
-/* array with IDs of enabled columns */
-static int columns[FINDMNT_NCOLUMNS];
-static int ncolumns;
 
 /* poll actions (parsed --poll=<list> */
 #define FINDMNT_NACTIONS	4		/* mount, umount, move, remount */
@@ -163,7 +186,7 @@ static int match_func(struct libmnt_fs *fs, void *data __attribute__ ((__unused_
 static int get_column_id(int num)
 {
 	assert(num < ncolumns);
-	assert(columns[num] < FINDMNT_NCOLUMNS);
+	assert((size_t) columns[num] < ARRAY_SIZE(infos));
 	return columns[num];
 }
 
@@ -174,7 +197,7 @@ static struct colinfo *get_column_info(int num)
 
 static const char *column_id_to_name(int id)
 {
-	assert(id < FINDMNT_NCOLUMNS);
+	assert((size_t) id < ARRAY_SIZE(infos));
 	return infos[id].name;
 }
 
@@ -195,25 +218,25 @@ static int get_column_flags(int num)
 
 static const char *get_match(int id)
 {
-	assert(id < FINDMNT_NCOLUMNS);
+	assert((size_t) id < ARRAY_SIZE(infos));
 	return infos[id].match;
 }
 
 static void *get_match_data(int id)
 {
-	assert(id < FINDMNT_NCOLUMNS);
+	assert((size_t) id < ARRAY_SIZE(infos));
 	return infos[id].match_data;
 }
 
 static void set_match(int id, const char *match)
 {
-	assert(id < FINDMNT_NCOLUMNS);
+	assert((size_t) id < ARRAY_SIZE(infos));
 	infos[id].match = match;
 }
 
 static void set_match_data(int id, void *data)
 {
-	assert(id < FINDMNT_NCOLUMNS);
+	assert((size_t) id < ARRAY_SIZE(infos));
 	infos[id].match_data = data;
 }
 
@@ -239,27 +262,34 @@ static void set_source_match(const char *data)
 static void enable_extra_target_match(void)
 {
 	char *cn = NULL, *mnt = NULL;
+	const char *tgt = NULL;
 
 	/*
 	 * Check if match pattern is mountpoint, if not use the
 	 * real mountpoint.
 	 */
-	cn = mnt_resolve_path(get_match(COL_TARGET), cache);
-	if (!cn)
-		return;
+	if (flags & FL_NOCACHE)
+		tgt = get_match(COL_TARGET);
+	else {
+		tgt = cn = mnt_resolve_path(get_match(COL_TARGET), cache);
+		if (!cn)
+			return;
+	}
 
-	mnt = mnt_get_mountpoint(cn);
-	if (!mnt || strcmp(mnt, cn) == 0)
-		return;
+	mnt = mnt_get_mountpoint(tgt);
+	if (mnt && strcmp(mnt, tgt) != 0)
+		set_match(COL_TARGET, mnt);	/* replace the current setting */
+	else
+		free(mnt);
 
-	/* replace the current setting with the real mountpoint */
-	set_match(COL_TARGET, mnt);
+	if (!cache)
+		free(cn);
 }
 
 
 static int is_tabdiff_column(int id)
 {
-	assert(id < FINDMNT_NCOLUMNS);
+	assert((size_t) id < ARRAY_SIZE(infos));
 
 	switch(id) {
 	case COL_ACTION:
@@ -339,10 +369,10 @@ static int is_mount_compatible_mode(void)
 
 static void disable_columns_truncate(void)
 {
-	int i;
+	size_t i;
 
-	for (i = 0; i < FINDMNT_NCOLUMNS; i++)
-		infos[i].flags &= ~TT_FL_TRUNC;
+	for (i = 0; i < ARRAY_SIZE(infos); i++)
+		infos[i].flags &= ~SCOLS_FL_TRUNC;
 }
 
 /*
@@ -350,9 +380,9 @@ static void disable_columns_truncate(void)
  */
 static int column_name_to_id(const char *name, size_t namesz)
 {
-	int i;
+	size_t i;
 
-	for (i = 0; i < FINDMNT_NCOLUMNS; i++) {
+	for (i = 0; i < ARRAY_SIZE(infos); i++) {
 		const char *cn = column_id_to_name(i);
 
 		if (!strncasecmp(name, cn, namesz) && !*(cn + namesz))
@@ -431,7 +461,7 @@ static char *get_tag(struct libmnt_fs *fs, const char *tagname, int col
 	else {
 		const char *dev = mnt_fs_get_source(fs);
 
-		if (dev)
+		if (dev && !(flags & FL_NOCACHE))
 			dev = mnt_resolve_spec(dev, cache);
 #ifdef HAVE_LIBUDEV
 		if (dev)
@@ -477,8 +507,14 @@ static char *get_vfs_attr(struct libmnt_fs *fs, int sizetype)
 		return sizestr;
 	}
 
-	return vfs_attr == 0 ? xstrdup("0") :
-		size_to_human_string(SIZE_SUFFIX_1LETTER, vfs_attr);
+	if (!vfs_attr)
+		sizestr = xstrdup("0");
+	else if (flags & FL_BYTES)
+		xasprintf(&sizestr, "%ju", vfs_attr);
+	else
+		sizestr = size_to_human_string(SIZE_SUFFIX_1LETTER, vfs_attr);
+
+	return sizestr;
 }
 
 /* reads FS data from libmount
@@ -493,19 +529,22 @@ static char *get_data(struct libmnt_fs *fs, int num)
 	{
 		const char *root = mnt_fs_get_root(fs);
 		const char *spec = mnt_fs_get_srcpath(fs);
+		char *cn = NULL;
 
 		if (spec && (flags & FL_CANONICALIZE))
-			spec = mnt_resolve_path(spec, cache);
+			spec = cn = mnt_resolve_path(spec, cache);
 		if (!spec) {
 			spec = mnt_fs_get_source(fs);
 
 			if (spec && (flags & FL_EVALUATE))
-				spec = mnt_resolve_spec(spec, cache);
+				spec = cn = mnt_resolve_spec(spec, cache);
 		}
 		if (root && spec && !(flags & FL_NOFSROOT) && strcmp(root, "/"))
 			xasprintf(&str, "%s[%s]", spec, root);
 		else if (spec)
 			str = xstrdup(spec);
+		if (!cache)
+			free(cn);
 		break;
 	}
 	case COL_TARGET:
@@ -545,7 +584,7 @@ static char *get_data(struct libmnt_fs *fs, int num)
 		if (!devno)
 			break;
 
-		if ((tt_flags & TT_FL_RAW) || (tt_flags & TT_FL_EXPORT))
+		if ((flags & FL_RAW) || (flags & FL_EXPORT))
 			xasprintf(&str, "%u:%u", major(devno), minor(devno));
 		else
 			xasprintf(&str, "%3u:%-3u", major(devno), minor(devno));
@@ -654,59 +693,68 @@ static char *get_tabdiff_data(struct libmnt_fs *old_fs,
 }
 
 /* adds one line to the output @tab */
-static struct tt_line *add_line(struct tt *tt, struct libmnt_fs *fs,
-					struct tt_line *parent)
+static struct libscols_line *add_line(struct libscols_table *table, struct libmnt_fs *fs,
+					struct libscols_line *parent)
 {
 	int i;
-	struct tt_line *line = tt_add_line(tt, parent);
+	struct libscols_line *line = scols_table_new_line(table, parent);
 
 	if (!line) {
 		warn(_("failed to add line to output"));
 		return NULL;
 	}
 	for (i = 0; i < ncolumns; i++)
-		tt_line_set_data(line, i, get_data(fs, i));
+		scols_line_refer_data(line, i, get_data(fs, i));
 
-	tt_line_set_userdata(line, fs);
+	scols_line_set_userdata(line, fs);
 	return line;
 }
 
-static struct tt_line *add_tabdiff_line(struct tt *tt, struct libmnt_fs *new_fs,
+static struct libscols_line *add_tabdiff_line(struct libscols_table *table, struct libmnt_fs *new_fs,
 			struct libmnt_fs *old_fs, int change)
 {
 	int i;
-	struct tt_line *line = tt_add_line(tt, NULL);
+	struct libscols_line *line = scols_table_new_line(table, NULL);
 
 	if (!line) {
 		warn(_("failed to add line to output"));
 		return NULL;
 	}
 	for (i = 0; i < ncolumns; i++)
-		tt_line_set_data(line, i,
+		scols_line_refer_data(line, i,
 				get_tabdiff_data(old_fs, new_fs, change, i));
 
 	return line;
 }
 
-static int has_line(struct tt *tt, struct libmnt_fs *fs)
+static int has_line(struct libscols_table *table, struct libmnt_fs *fs)
 {
-	struct list_head *p;
+	struct libscols_line *ln;
+	struct libscols_iter *itr;
+	int rc = 0;
 
-	list_for_each(p, &tt->tb_lines) {
-		struct tt_line *ln = list_entry(p, struct tt_line, ln_lines);
-		if ((struct libmnt_fs *) ln->userdata == fs)
-			return 1;
+	itr = scols_new_iter(SCOLS_ITER_FORWARD);
+	if (!itr)
+		return 0;
+
+	while (scols_table_next_line(table, itr, &ln) == 0) {
+		if ((struct libmnt_fs *) scols_line_get_userdata(ln) == fs) {
+			rc = 1;
+			break;
+		}
 	}
-	return 0;
+
+	scols_free_iter(itr);
+	return rc;
 }
 
-/* reads filesystems from @tb (libmount) and fillin @tt (output table) */
-static int create_treenode(struct tt *tt, struct libmnt_table *tb,
-			   struct libmnt_fs *fs, struct tt_line *parent_line)
+/* reads filesystems from @tb (libmount) and fillin @table (output table) */
+static int create_treenode(struct libscols_table *table, struct libmnt_table *tb,
+			   struct libmnt_fs *fs, struct libscols_line *parent_line)
 {
 	struct libmnt_fs *chld = NULL;
 	struct libmnt_iter *itr = NULL;
-	struct tt_line *line;
+	struct libscols_line *line;
 	int rc = -1;
 
 	if (!fs) {
@@ -715,7 +763,7 @@ static int create_treenode(struct tt *tt, struct libmnt_table *tb,
 			goto leave;
 		parent_line = NULL;
 
-	} else if ((flags & FL_SUBMOUNTS) && has_line(tt, fs))
+	} else if ((flags & FL_SUBMOUNTS) && has_line(table, fs))
 		return 0;
 
 	itr = mnt_new_iter(MNT_ITER_FORWARD);
@@ -723,7 +771,7 @@ static int create_treenode(struct tt *tt, struct libmnt_table *tb,
 		goto leave;
 
 	if ((flags & FL_SUBMOUNTS) || match_func(fs, NULL)) {
-		line = add_line(tt, fs, parent_line);
+		line = add_line(table, fs, parent_line);
 		if (!line)
 			goto leave;
 	} else
@@ -733,7 +781,7 @@ static int create_treenode(struct tt *tt, struct libmnt_table *tb,
 	 * add all children to the output table
 	 */
 	while(mnt_table_next_child_fs(tb, itr, fs, &chld) == 0) {
-		if (create_treenode(tt, tb, chld, line))
+		if (create_treenode(table, tb, chld, line))
 			goto leave;
 	}
 	rc = 0;
@@ -810,6 +858,29 @@ static struct libmnt_table *parse_tabfiles(char **files,
 	return tb;
 }
 
+/*
+ * Parses mountinfo and calls mnt_cache_set_targets(cache, mtab). Only
+ * necessary if @tb in main() was read from a non-kernel source.
+ */
+static void cache_set_targets(struct libmnt_cache *tmp)
+{
+	struct libmnt_table *tb;
+	const char *path;
+
+	tb = mnt_new_table();
+	if (!tb)
+		return;
+
+	path = access(_PATH_PROC_MOUNTINFO, R_OK) == 0 ?
+		_PATH_PROC_MOUNTINFO :
+		_PATH_PROC_MOUNTS;
+
+	if (mnt_table_parse_file(tb, path) == 0)
+		mnt_cache_set_targets(tmp, tb);
+
+	mnt_unref_table(tb);
+}
+
 /* checks if @tb contains parent->child relations */
 static int tab_is_tree(struct libmnt_table *tb)
 {
@@ -829,6 +900,26 @@ static int tab_is_tree(struct libmnt_table *tb)
 	return rc;
 }
 
+/* checks if all fs in @tb are from kernel */
+static int tab_is_kernel(struct libmnt_table *tb)
+{
+	struct libmnt_fs *fs = NULL;
+	struct libmnt_iter *itr = NULL;
+
+	itr = mnt_new_iter(MNT_ITER_BACKWARD);
+	if (!itr)
+		return 0;
+
+	while (mnt_table_next_fs(tb, itr, &fs) == 0) {
+		if (!mnt_fs_is_kernel(fs)) {
+			mnt_free_iter(itr);
+			return 0;
+		}
+	}
+
+	mnt_free_iter(itr);
+	return 1;
+}
 
 /* filter function for libmount (mnt_table_find_next_fs()) */
 static int match_func(struct libmnt_fs *fs,
@@ -927,7 +1018,7 @@ again:
  * --submounts tree output.
  */
 static int add_matching_lines(struct libmnt_table *tb,
-			      struct tt *tt, int direction)
+			      struct libscols_table *table, int direction)
 {
 	struct libmnt_iter *itr = NULL;
 	struct libmnt_fs *fs;
@@ -940,10 +1031,10 @@ static int add_matching_lines(struct libmnt_table *tb,
 	}
 
 	while((fs = get_next_fs(tb, itr))) {
-		if ((tt_flags & TT_FL_TREE) || (flags & FL_SUBMOUNTS))
-			rc = create_treenode(tt, tb, fs, NULL);
+		if ((flags & FL_TREE) || (flags & FL_SUBMOUNTS))
+			rc = create_treenode(table, tb, fs, NULL);
 		else
-			rc = !add_line(tt, fs, NULL);
+			rc = !add_line(table, fs, NULL);
 		if (rc)
 			goto done;
 		nlines++;
@@ -984,7 +1075,7 @@ static int poll_match(struct libmnt_fs *fs)
 }
 
 static int poll_table(struct libmnt_table *tb, const char *tabfile,
-		  int timeout, struct tt *tt, int direction)
+		  int timeout, struct libscols_table *table, int direction)
 {
 	FILE *f = NULL;
 	int rc = -1;
@@ -1056,7 +1147,7 @@ static int poll_table(struct libmnt_table *tb, const char *tabfile,
 			if (!poll_match(new ? new : old))
 				continue;
 			count++;
-			rc = !add_tabdiff_line(tt, new, old, change);
+			rc = !add_tabdiff_line(table, new, old, change);
 			if (rc)
 				goto done;
 			if (flags & FL_FIRSTONLY)
@@ -1064,7 +1155,7 @@ static int poll_table(struct libmnt_table *tb, const char *tabfile,
 		}
 
 		if (count) {
-			rc = tt_print_table(tt);
+			rc = scols_print_table(table);
 			if (rc)
 				goto done;
 		}
@@ -1074,7 +1165,7 @@ static int poll_table(struct libmnt_table *tb, const char *tabfile,
 		tb = tb_new;
 		tb_new = tmp;
 
-		tt_remove_lines(tt);
+		scols_table_remove_lines(table);
 		mnt_reset_table(tb_new);
 
 		if (count && (flags & FL_FIRSTONLY))
@@ -1091,9 +1182,17 @@ done:
 	return rc;
 }
 
+static int uniq_fs_target_cmp(
+		struct libmnt_table *tb __attribute__((__unused__)),
+		struct libmnt_fs *a,
+		struct libmnt_fs *b)
+{
+	return !mnt_fs_match_target(a, mnt_fs_get_target(b), cache);
+}
+
 static void __attribute__((__noreturn__)) usage(FILE *out)
 {
-	int i;
+	size_t i;
 
 	fputs(USAGE_HEADER, out);
 	fprintf(out, _(
@@ -1103,45 +1202,46 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	" %1$s [options] [--source <device>] [--target <mountpoint>]\n"),
 		program_invocation_short_name);
 
-	fprintf(out, _(
-	"\nOptions:\n"
-	" -s, --fstab            search in static table of filesystems\n"
-	" -m, --mtab             search in table of mounted filesystems\n"
-	" -k, --kernel           search in kernel table of mounted\n"
-        "                          filesystems (default)\n\n"));
+	fputs(USAGE_SEPARATOR, out);
+	fputs(_("Find a (mounted) filesystem.\n"), out);
 
-	fprintf(out, _(
-	" -p, --poll[=<list>]    monitor changes in table of mounted filesystems\n"
-	" -w, --timeout <num>    upper limit in milliseconds that --poll will block\n\n"));
+	fputs(USAGE_OPTIONS, out);
+	fputs(_(" -s, --fstab            search in static table of filesystems\n"), out);
+	fputs(_(" -m, --mtab             search in table of mounted filesystems\n"), out);
+	fputs(_(" -k, --kernel           search in kernel table of mounted\n"
+		"                          filesystems (default)\n"), out);
+	fputc('\n', out);
+	fputs(_(" -p, --poll[=<list>]    monitor changes in table of mounted filesystems\n"), out);
+	fputs(_(" -w, --timeout <num>    upper limit in milliseconds that --poll will block\n"), out);
+	fputc('\n', out);
 
-	fprintf(out, _(
-	" -A, --all              disable all built-in filters, print all filesystems\n"
-	" -a, --ascii            use ASCII chars for tree formatting\n"
-	" -c, --canonicalize     canonicalize printed paths\n"
-	" -D, --df               imitate the output of df(1)\n"
-	" -d, --direction <word> direction of search, 'forward' or 'backward'\n"
-	" -e, --evaluate         convert tags (LABEL,UUID,PARTUUID,PARTLABEL) \n"
-	"                          to device names\n"
-	" -F, --tab-file <path>  alternative file for --fstab, --mtab or --kernel options\n"
-	" -f, --first-only       print the first found filesystem only\n"));
-
-	fprintf(out, _(
-	" -i, --invert           invert the sense of matching\n"
-	" -l, --list             use list format output\n"
-	" -N, --task <tid>       use alternative namespace (/proc/<tid>/mountinfo file)\n"
-	" -n, --noheadings       don't print column headings\n"
-	" -u, --notruncate       don't truncate text in columns\n"));
-fputs (_(" -O, --options <list>   limit the set of filesystems by mount options\n"), out);
-fputs (_(" -o, --output <list>    the output columns to be shown\n"), out);
-fputs (_(" -P, --pairs            use key=\"value\" output format\n"), out);
-fputs (_(" -r, --raw              use raw output format\n"), out);
-fputs (_(" -t, --types <list>     limit the set of filesystems by FS types\n"), out);
-	fprintf(out, _(
-	" -v, --nofsroot         don't print [/dir] for bind or btrfs mounts\n"
-	" -R, --submounts        print all submounts for the matching filesystems\n"
-	" -S, --source <string>  the device to mount (by name, maj:min, \n"
-	"                          LABEL=, UUID=, PARTUUID=, PARTLABEL=)\n"
-	" -T, --target <string>  the mountpoint to use\n"));
+	fputs(_(" -A, --all              disable all built-in filters, print all filesystems\n"), out);
+	fputs(_(" -a, --ascii            use ASCII chars for tree formatting\n"), out);
+	fputs(_(" -b, --bytes            print sizes in bytes rather than in human readable format\n"), out);
+	fputs(_(" -C, --nocanonicalize   don't canonicalize when comparing paths\n"), out);
+	fputs(_(" -c, --canonicalize     canonicalize printed paths\n"), out);
+	fputs(_(" -D, --df               imitate the output of df(1)\n"), out);
+	fputs(_(" -d, --direction <word> direction of search, 'forward' or 'backward'\n"), out);
+	fputs(_(" -e, --evaluate         convert tags (LABEL,UUID,PARTUUID,PARTLABEL) \n"
+	        "                          to device names\n"), out);
+	fputs(_(" -F, --tab-file <path>  alternative file for -s, -m or -k options\n"), out);
+	fputs(_(" -f, --first-only       print the first found filesystem only\n"), out);
+	fputs(_(" -i, --invert           invert the sense of matching\n"), out);
+	fputs(_(" -l, --list             use list format output\n"), out);
+	fputs(_(" -N, --task <tid>       use alternative namespace (/proc/<tid>/mountinfo file)\n"), out);
+	fputs(_(" -n, --noheadings       don't print column headings\n"), out);
+	fputs(_(" -O, --options <list>   limit the set of filesystems by mount options\n"), out);
+	fputs(_(" -o, --output <list>    the output columns to be shown\n"), out);
+	fputs(_(" -P, --pairs            use key=\"value\" output format\n"), out);
+	fputs(_(" -R, --submounts        print all submounts for the matching filesystems\n"), out);
+	fputs(_(" -r, --raw              use raw output format\n"), out);
+	fputs(_(" -S, --source <string>  the device to mount (by name, maj:min, \n"
+	        "                          LABEL=, UUID=, PARTUUID=, PARTLABEL=)\n"), out);
+	fputs(_(" -T, --target <string>  the mountpoint to use\n"), out);
+	fputs(_(" -t, --types <list>     limit the set of filesystems by FS types\n"), out);
+	fputs(_(" -U, --uniq             ignore filesystems with duplicate target\n"), out);
+	fputs(_(" -u, --notruncate       don't truncate text in columns\n"), out);
+	fputs(_(" -v, --nofsroot         don't print [/dir] for bind or btrfs mounts\n"), out);
 
 	fputs(USAGE_SEPARATOR, out);
 	fputs(USAGE_HELP, out);
@@ -1149,7 +1249,7 @@ fputs (_(" -t, --types <list>     limit the set of filesystems by FS types\n"),
 
 	fprintf(out, _("\nAvailable columns:\n"));
 
-	for (i = 0; i < FINDMNT_NCOLUMNS; i++)
+	for (i = 0; i < ARRAY_SIZE(infos); i++)
 		fprintf(out, " %11s  %s\n", infos[i].name, _(infos[i].help));
 
 	fprintf(out, USAGE_MAN_TAIL("findmnt(8)"));
@@ -1166,11 +1266,12 @@ int main(int argc, char *argv[])
 	int ntabfiles = 0, tabtype = 0;
 	char *outarg = NULL;
 
-	struct tt *tt = NULL;
+	struct libscols_table *table = NULL;
 
 	static const struct option longopts[] = {
 	    { "all",          0, 0, 'A' },
 	    { "ascii",        0, 0, 'a' },
+	    { "bytes",        0, 0, 'b' },
 	    { "canonicalize", 0, 0, 'c' },
 	    { "direction",    1, 0, 'd' },
 	    { "df",           0, 0, 'D' },
@@ -1190,6 +1291,7 @@ int main(int argc, char *argv[])
 	    { "pairs",        0, 0, 'P' },
 	    { "raw",          0, 0, 'r' },
 	    { "types",        1, 0, 't' },
+	    { "nocanonicalize", 0, 0, 'C' },
 	    { "nofsroot",     0, 0, 'v' },
 	    { "submounts",    0, 0, 'R' },
 	    { "source",       1, 0, 'S' },
@@ -1197,12 +1299,15 @@ int main(int argc, char *argv[])
 	    { "task",         1, 0, 'N' },
 	    { "target",       1, 0, 'T' },
 	    { "timeout",      1, 0, 'w' },
+	    { "uniq",         0, 0, 'U' },
 	    { "version",      0, 0, 'V' },
 
 	    { NULL,           0, 0, 0 }
 	};
 
 	static const ul_excl_t excl[] = {	/* rows and cols in in ASCII order */
+		{ 'C', 'c'},                    /* [no]canonicalize */
+		{ 'C', 'e' },			/* nocanonicalize, evaluate */
 		{ 'N','k','m','s' },		/* task,kernel,mtab,fstab */
 		{ 'P','l','r' },		/* pairs,list,raw */
 		{ 'm','p','s' },		/* mtab,poll,fstab */
@@ -1210,18 +1315,16 @@ int main(int argc, char *argv[])
 	};
 	int excl_st[ARRAY_SIZE(excl)] = UL_EXCL_STATUS_INIT;
 
-	assert(ARRAY_SIZE(columns) == FINDMNT_NCOLUMNS);
-
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 	atexit(close_stdout);
 
 	/* default output format */
-	tt_flags |= TT_FL_TREE;
+	flags |= FL_TREE;
 
 	while ((c = getopt_long(argc, argv,
-				"AacDd:ehifF:o:O:p::PklmnN:rst:uvRS:T:w:V",
+				"AabCcDd:ehifF:o:O:p::PklmnN:rst:uvRS:T:Uw:V",
 				longopts, NULL)) != -1) {
 
 		err_exclusive_options(c, longopts, excl, excl_st);
@@ -1231,13 +1334,19 @@ int main(int argc, char *argv[])
 			flags |= FL_ALL;
 			break;
 		case 'a':
-			tt_flags |= TT_FL_ASCII;
+			flags |= FL_ASCII;
+			break;
+		case 'b':
+			flags |= FL_BYTES;
+			break;
+		case 'C':
+			flags |= FL_NOCACHE;
 			break;
 		case 'c':
 			flags |= FL_CANONICALIZE;
 			break;
 		case 'D':
-			tt_flags &= ~TT_FL_TREE;
+			flags &= ~FL_TREE;
 			flags |= FL_DF;
 			break;
 		case 'd':
@@ -1282,19 +1391,19 @@ int main(int argc, char *argv[])
 					exit(EXIT_FAILURE);
 			}
 			flags |= FL_POLL;
-			tt_flags &= ~TT_FL_TREE;
+			flags &= ~FL_TREE;
 			break;
 		case 'P':
-			tt_flags |= TT_FL_EXPORT;
-			tt_flags &= ~TT_FL_TREE;
+			flags |= FL_EXPORT;
+			flags &= ~FL_TREE;
 			break;
 		case 'm':		/* mtab */
 			tabtype = TABTYPE_MTAB;
-			tt_flags &= ~TT_FL_TREE;
+			flags &= ~FL_TREE;
 			break;
 		case 's':		/* fstab */
 			tabtype = TABTYPE_FSTAB;
-			tt_flags &= ~TT_FL_TREE;
+			flags &= ~FL_TREE;
 			break;
 		case 'k':		/* kernel (mountinfo) */
 			tabtype = TABTYPE_KERNEL;
@@ -1303,14 +1412,14 @@ int main(int argc, char *argv[])
 			set_match(COL_FSTYPE, optarg);
 			break;
 		case 'r':
-			tt_flags &= ~TT_FL_TREE;	/* disable the default */
-			tt_flags |= TT_FL_RAW;		/* enable raw */
+			flags &= ~FL_TREE;	/* disable the default */
+			flags |= FL_RAW;	/* enable raw */
 			break;
 		case 'l':
-			tt_flags &= ~TT_FL_TREE; /* disable the default */
+			flags &= ~FL_TREE;	/* disable the default */
 			break;
 		case 'n':
-			tt_flags |= TT_FL_NOHEADINGS;
+			flags |= FL_NOHEADINGS;
 			break;
 		case 'N':
 			tabtype = TABTYPE_KERNEL;
@@ -1332,6 +1441,9 @@ int main(int argc, char *argv[])
 			set_match(COL_TARGET, optarg);
 			flags |= FL_NOSWAPMATCH;
 			break;
+		case 'U':
+			flags |= FL_UNIQ;
+			break;
 		case 'w':
 			timeout = strtos32_or_err(optarg, _("invalid timeout argument"));
 			break;
@@ -1345,24 +1457,24 @@ int main(int argc, char *argv[])
 	}
 
 	if (!ncolumns && (flags & FL_DF)) {
-		columns[ncolumns++] = COL_SOURCE;
-		columns[ncolumns++] = COL_FSTYPE;
-		columns[ncolumns++] = COL_SIZE;
-		columns[ncolumns++] = COL_USED;
-		columns[ncolumns++] = COL_AVAIL;
-		columns[ncolumns++] = COL_USEPERC;
-		columns[ncolumns++] = COL_TARGET;
+		add_column(columns, ncolumns++, COL_SOURCE);
+		add_column(columns, ncolumns++, COL_FSTYPE);
+		add_column(columns, ncolumns++, COL_SIZE);
+		add_column(columns, ncolumns++, COL_USED);
+		add_column(columns, ncolumns++, COL_AVAIL);
+		add_column(columns, ncolumns++, COL_USEPERC);
+		add_column(columns, ncolumns++, COL_TARGET);
 	}
 
 	/* default columns */
 	if (!ncolumns) {
 		if (flags & FL_POLL)
-			columns[ncolumns++] = COL_ACTION;
+			add_column(columns, ncolumns++, COL_ACTION);
 
-		columns[ncolumns++] = COL_TARGET;
-		columns[ncolumns++] = COL_SOURCE;
-		columns[ncolumns++] = COL_FSTYPE;
-		columns[ncolumns++] = COL_OPTIONS;
+		add_column(columns, ncolumns++, COL_TARGET);
+		add_column(columns, ncolumns++, COL_SOURCE);
+		add_column(columns, ncolumns++, COL_FSTYPE);
+		add_column(columns, ncolumns++, COL_OPTIONS);
 	}
 
 	if (outarg && string_add_to_idarray(outarg, columns, ARRAY_SIZE(columns),
@@ -1393,7 +1505,7 @@ int main(int argc, char *argv[])
 	    || get_match(COL_TARGET)
 	    || get_match(COL_SOURCE)
 	    || get_match(COL_MAJMIN)))
-		tt_flags &= ~TT_FL_TREE;
+		flags &= ~FL_TREE;
 
 	if (!(flags & FL_NOSWAPMATCH) &&
 	    !get_match(COL_TARGET) && get_match(COL_SOURCE)) {
@@ -1412,44 +1524,59 @@ int main(int argc, char *argv[])
 	 * initialize libmount
 	 */
 	mnt_init_debug(0);
+	scols_init_debug(0);
 
 	tb = parse_tabfiles(tabfiles, ntabfiles, tabtype);
 	if (!tb)
 		goto leave;
 
-	if ((tt_flags & TT_FL_TREE) && (ntabfiles > 1 || !tab_is_tree(tb)))
-		tt_flags &= ~TT_FL_TREE;
+	if (tabtype == TABTYPE_MTAB && tab_is_kernel(tb))
+		tabtype = TABTYPE_KERNEL;
 
-	cache = mnt_new_cache();
-	if (!cache) {
-		warn(_("failed to initialize libmount cache"));
-		goto leave;
+	if ((flags & FL_TREE) && (ntabfiles > 1 || !tab_is_tree(tb)))
+		flags &= ~FL_TREE;
+
+	if (!(flags & FL_NOCACHE)) {
+		cache = mnt_new_cache();
+		if (!cache) {
+			warn(_("failed to initialize libmount cache"));
+			goto leave;
+		}
+		mnt_table_set_cache(tb, cache);
+
+		if (tabtype != TABTYPE_KERNEL)
+			cache_set_targets(cache);
 	}
-	mnt_table_set_cache(tb, cache);
 
+	if (flags & FL_UNIQ)
+		mnt_table_uniq_fs(tb, MNT_UNIQ_KEEPTREE, uniq_fs_target_cmp);
 
 	/*
-	 * initialize output formatting (tt.h)
+	 * initialize output formatting (libsmartcols.h)
 	 */
-	tt = tt_new_table(tt_flags | TT_FL_FREEDATA);
-	if (!tt) {
+	table = scols_new_table();
+	if (!table) {
 		warn(_("failed to initialize output table"));
 		goto leave;
 	}
+	scols_table_enable_raw(table,        !!(flags & FL_RAW));
+	scols_table_enable_export(table,     !!(flags & FL_EXPORT));
+	scols_table_enable_ascii(table,      !!(flags & FL_ASCII));
+	scols_table_enable_noheadings(table, !!(flags & FL_NOHEADINGS));
 
 	for (i = 0; i < ncolumns; i++) {
 		int fl = get_column_flags(i);
 		int id = get_column_id(i);
 
-		if (!(tt_flags & TT_FL_TREE))
-			fl &= ~TT_FL_TREE;
+		if (!(flags & FL_TREE))
+			fl &= ~SCOLS_FL_TREE;
 
 		if (!(flags & FL_POLL) && is_tabdiff_column(id)) {
 			warnx(_("%s column is requested, but --poll "
 			       "is not enabled"), get_column_name(i));
 			goto leave;
 		}
-		if (!tt_define_column(tt, get_column_name(i),
+		if (!scols_table_new_column(table, get_column_name(i),
 					get_column_whint(i), fl)) {
 			warn(_("failed to initialize output column"));
 			goto leave;
@@ -1461,14 +1588,14 @@ int main(int argc, char *argv[])
 	 */
 	if (flags & FL_POLL) {
 		/* poll mode (accept the first tabfile only) */
-		rc = poll_table(tb, tabfiles ? *tabfiles : _PATH_PROC_MOUNTINFO, timeout, tt, direction);
+		rc = poll_table(tb, tabfiles ? *tabfiles : _PATH_PROC_MOUNTINFO, timeout, table, direction);
 
-	} else if ((tt_flags & TT_FL_TREE) && !(flags & FL_SUBMOUNTS)) {
+	} else if ((flags & FL_TREE) && !(flags & FL_SUBMOUNTS)) {
 		/* whole tree */
-		rc = create_treenode(tt, tb, NULL, NULL);
+		rc = create_treenode(table, tb, NULL, NULL);
 	} else {
 		/* whole lits of sub-tree */
-		rc = add_matching_lines(tb, tt, direction);
+		rc = add_matching_lines(tb, table, direction);
 
 		if (rc != 0
 		    && tabtype == TABTYPE_KERNEL
@@ -1480,7 +1607,7 @@ int main(int argc, char *argv[])
 			 * match
 			 */
 			enable_extra_target_match();
-			rc = add_matching_lines(tb, tt, direction);
+			rc = add_matching_lines(tb, table, direction);
 		}
 	}
 
@@ -1488,9 +1615,9 @@ int main(int argc, char *argv[])
 	 * Print the output table for non-poll modes
 	 */
 	if (!rc && !(flags & FL_POLL))
-		tt_print_table(tt);
+		scols_print_table(table);
 leave:
-	tt_free_table(tt);
+	scols_unref_table(table);
 
 	mnt_unref_table(tb);
 	mnt_unref_cache(cache);
